package main

import (
	"context"
	"flag"
	"fmt"
	"io"
	"log"
	"net/http"
	"os"
	"os/signal"
	"sync"

	"github.com/r3labs/sse"
)

var (
	endpoint string
	host     string
	port     int
	clients  int
)

func init() {
	flag.StringVar(&endpoint, "endpoint", "stream", "path to hit on the server")
	flag.StringVar(&host, "host", "localhost", "host that the server is running on")
	flag.IntVar(&port, "port", 7000, "port that the server is running on")
	flag.IntVar(&clients, "clients", 1, "number of clients to connect to the server")
	flag.Parse()
}

func main() {
	log.SetFlags(log.Lshortfile)

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	sigc := make(chan os.Signal, 1)
	signal.Notify(sigc, os.Interrupt)
	go func() {
		<-sigc
		cancel()
	}()

	url := fmt.Sprintf("http://%s:%d/%s", host, port, endpoint)
	cs := make([]*Client, clients)

	for i := 0; i < clients; i++ {
		cs[i] = NewClient(i, url)
	}

	wg := sync.WaitGroup{}
	for _, client := range cs {
		wg.Add(1)

		go func(c *Client) {
			defer wg.Done()
			c.makeRequest(ctx)
		}(client)
	}

	fmt.Println("waiting")
	wg.Wait()
	fmt.Println("done")
}

type Client struct {
	id  int
	url string
}

func NewClient(id int, url string) *Client {
	return &Client{
		id:  id,
		url: url,
	}
}

func (c *Client) makeRequest(ctx context.Context) error {
	req, err := http.NewRequestWithContext(ctx, http.MethodGet, c.url, nil)
	if err != nil {
		return fmt.Errorf("failed to create request: %s", err)
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return fmt.Errorf("failed to make request: %s", err)
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf("non 200 status code, code=%d", resp.StatusCode)
	}

	reader := sse.NewEventStreamReader(resp.Body)

	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		default:
			log.Println("waiting for event")
			b, err := reader.ReadEvent()
			if err != nil {
				if err == io.EOF {
					log.Println("Got EOF")
					return nil
				}
				log.Println("err reading event: ", err)
			}
			log.Printf("client %d got event: %s \n", c.id, string(b))
		}
	}
}
