package main

import (
	"bytes"
	"context"
	"flag"
	"fmt"
	"log"
	"net/http"
	"os"
	"os/signal"
	"sync"
	"time"
)

var (
	concurrency int
	streamName  string
	duration    int
	rps         int
)

func init() {
	flag.IntVar(&concurrency, "concurrency", 1, "the number of concurrent publishers")
	flag.StringVar(&streamName, "stream-name", "james-test", "the pushpin channel to publish events to")
	flag.IntVar(&duration, "duration", 60, "time in seconds to run the publishers for")
	flag.IntVar(&rps, "rps", 100, "max number of messages each publishers send per second")

	flag.Parse()
}

func main() {
	log.Println("running publisher...", "concurrency", concurrency, "duration", duration, "rps", rps, "streamName", streamName)
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	sigc := make(chan os.Signal, 1)
	signal.Notify(sigc, os.Interrupt)
	go func() {
		<-sigc
		cancel()
	}()

	publishers := make([]*Publisher, concurrency)
	for i := 0; i < concurrency; i++ {
		publishers[i] = NewPublisher(streamName, rps, duration)
	}

	wg := sync.WaitGroup{}
	for _, p := range publishers {
		wg.Add(1)

		go func(pub *Publisher) {
			defer wg.Done()

			if err := pub.Publish(ctx); err != nil {
				log.Println("Publisher failed: ", err)
			}
		}(p)
	}

	fmt.Println("waiting...")
	wg.Wait()
	fmt.Println("done")
}

type Publisher struct {
	rps      int
	duration time.Duration
	channel  string
}

func NewPublisher(channel string, rps int, duration int) *Publisher {
	return &Publisher{
		rps:      rps,
		duration: time.Duration(duration) * time.Second,
		channel:  channel,
	}
}

func (p *Publisher) Publish(ctx context.Context) error {
	messageC := p.messageGenerator(ctx)
	throttle := time.NewTicker(time.Duration(1e6/(p.rps)) * time.Microsecond)
	defer func() {
		throttle.Stop()
		p.closeStream()
	}()

	for {
		select {
		case <-ctx.Done():
			return ctx.Err()
		case message, ok := <-messageC:
			if !ok {
				return nil
			}

			<-throttle.C
			if err := p.publishMessage(ctx, message); err != nil {
				log.Println("failed to publish message: ", err)
			}
		}
	}
}

func (p *Publisher) publishMessage(ctx context.Context, message []byte) error {
	r, err := http.NewRequestWithContext(ctx, http.MethodPost, "http://localhost:5561/publish", bytes.NewBuffer(message))
	if err != nil {
		return err
	}

	resp, err := http.DefaultClient.Do(r)
	if err != nil {
		return err
	}

	if resp.StatusCode != http.StatusOK {
		return fmt.Errorf("non 200 status code: %d", resp.StatusCode)
	}

	return nil
}

// messageGenerator generates messages for the duration that the publisher
// is meant to run for
func (p *Publisher) messageGenerator(ctx context.Context) <-chan []byte {
	out := make(chan []byte)

	go func() {
		defer close(out)

		start := time.Now()
		idx := 0
		body := `{ "items": [ { "channel": "%s", "formats": {
"http-stream": { "content": "hello there event: %d\n" } } } ] }`

		for time.Since(start) < p.duration {
			select {
			case <-ctx.Done():
				return
			case out <- []byte(fmt.Sprintf(body, p.channel, idx)):
				idx++
			}
		}
	}()

	return out
}

func (p *Publisher) closeStream() {

	message := `
		{
  "items": [
    {
      "channel": "%s",
      "id": "an-item-id",
      "formats": {
        "http-stream": {
          "action": "close"
        }
      }
    }
  ]
}`

	b := []byte(fmt.Sprintf(message, p.channel))

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	log.Println("closing channel")
	p.publishMessage(ctx, b)
}
